/***************************************************************************
    HTTP Sample (Operating as command in shell)
    Copyright (c)  2014-2017, eForce Co., Ltd. All rights reserved.

    2014/04/02: Created.
    2017/01/01: Changed content reception processing.
    2017/09/04: Changed Chunked receive function.
                  - remove CRLF.
                  - receive CRLF after receive chunked size 0 (last-chunked)
                Changed return value check http_rcv_res().
                  treat as error when http_rcv_res() return value is not
                  equal argument.
    2017/09/05: Changed configuration string SSL_SUP -> HTTPC_SSL_SUP
                Remove compile error codes, when SSL enabled.
 ***************************************************************************/
#include "kernel.h"
#include "sample_netapp_cfg.h"      /* include Sample Application Settings */
#include "http_client.h"

static VB http_buf[1024];

#ifdef ENA_AUTH_BASIC
static VB basic_user[SPL_HTTP_USR_BUF]  = SPL_HTTP_B_USR;
static VB basic_pass[SPL_HTTP_PW_BUF]   = SPL_HTTP_B_PW;
#endif
#ifdef ENA_AUTH_DIGEST
static VB digest_user[SPL_HTTP_USR_BUF] = SPL_HTTP_D_USR;
static VB digest_pass[SPL_HTTP_PW_BUF]  = SPL_HTTP_D_PW;
#endif
static UB use_proxy = 0;
static UW prx_ipa;
static UW prx_port;

#ifdef HTTPC_SSL_SUP
/* printf like function */
extern void shell_printf(VP ctrl, const char *fmt, ...) ;
#endif

/* Receive chunk contents */
static ER rcv_chunked_contents(VP ctrl, T_HTTP_CLIENT *http)
{
    UB chunkhead;
    UH chunklen;
    ER ercd;
    int len;
    UH l;

    while (1) {
        /* read chunk head */
        chunklen = 0;
        while (1) {
            ercd = http_rcv_res(http, &chunkhead, 1);
            if (ercd != 1) {
                shell_puts(ctrl, "HTTP response error"SPL_LF);
                return ercd;
            }
            if (chunkhead == 0x0D) {
                continue;
            }
            if (chunkhead == 0x0A) {
                break;
            }
            if ('0' <= chunkhead && chunkhead <= '9') {
                chunkhead -= '0';
            } else if ('a' <= chunkhead && chunkhead <= 'f') {
                chunkhead -= 87;
            } else if ('A' <= chunkhead && chunkhead <= 'F') {
                chunkhead -= 65;
            } else {
                shell_puts(ctrl, "Recognize chunk header error"SPL_LF);
                return E_OBJ;
            }
            chunklen <<= 4;
            chunklen += chunkhead;
        }
        
        /* last-chunk */
        if (chunklen == 0) {
            break;
        }
        
        chunklen += 2;  /* +2 : length of CRLF */

        /* read data */
        while (chunklen > 0) {
            net_memset(http_buf, 0, sizeof(http_buf));
            len = sizeof(http_buf);
            if (len > chunklen) {
                len = chunklen;
            }
            ercd = http_rcv_res(http, (UB*)http_buf, len);
            if (ercd != len) {
                if (ercd < 0) {
                    return ercd;
                } else {
                    return E_TMOUT;
                }
            }
            
            /* remove CRLF (bounday of chunk-data) */
            l = chunklen - ercd;
            if (l == 1) {
                http_buf[ercd - 1] = '\0'; /* CR -> '\0' */
            }
            if (l == 0) {
                if (ercd == 1) {
                    http_buf[0] = '0';         /* LF -> '\0' */
                } else {
                    http_buf[ercd - 2] = '\0'; /* CR -> '\0' */
                    http_buf[ercd - 1] = '\0'; /* LF -> '\0' */
                }
            }
            
            shell_puts(ctrl, http_buf);
            
            chunklen -= ercd;
        }
    }
    
    /* receive CRLF (end of chunk-body) */
    ercd = http_rcv_res(http, (UB *)http_buf, 2);
    if (ercd != 2) {
        return E_TMOUT;
    }
    
    if ((http_buf[0] != 0x0d) || (http_buf[1] != 0x0a)) {
        return E_OBJ;
    }
    
    return E_OK;
}

/* Receive contents */
static ER rcv_contents(VP ctrl, T_HTTP_CLIENT *http)
{
    ER ercd;
    int len, i;

    i = 0;
    while (i < http->response.contentlen) {
        net_memset(http_buf, 0, sizeof(http_buf));
        len = sizeof(http_buf);
        len = len < http->response.contentlen-i ? len : http->response.contentlen-i;

        ercd = http_rcv_res(http, (UB*)http_buf, len);
        if (ercd <= 0) {
            break;
        }
        shell_puts(ctrl, http_buf);
        i += ercd;
    }
    return E_OK;
}

/* Receiving contents until it is disconnected from the server side */
static ER rcv_continuous_contents(VP ctrl, T_HTTP_CLIENT *http)
{
    ER ercd;

    while (1) {
        net_memset(http_buf, 0, sizeof(http_buf));
        ercd = http_rcv_res(http, (UB*)http_buf, sizeof(http_buf));
        if (ercd != sizeof(http_buf)) {
            break;
        }
        shell_puts(ctrl, http_buf);
    }
    return E_OK;
}

static ER rcv_contents_mng(VP ctrl, T_HTTP_CLIENT *http)
{
    ER ercd;
    
    /* chunk`*/
    if (http->response.chunked) {
        ercd = rcv_chunked_contents(ctrl, http);
    } else if (http->response.contentlen){
        ercd = rcv_contents(ctrl, http);
    } else {
        /* streaming ?*/
        ercd = rcv_continuous_contents(ctrl, http);
    }
    
    return ercd;
}

#ifdef HTTPC_SSL_SUP
/* Output SSL Infomation */
void shell_puts_SSL(VP ctrl, ID ssl_id)
{
    ER ercd;
    UB num;
    T_VALID_TIM tim;
    VB buf[128];

    /* Certificate information */
    shell_puts(ctrl, SPL_LF SPL_LF "-- Certificate --" SPL_LF);

    /* Subject */
    shell_puts(ctrl, SPL_LF"  [Subject] ");
    ercd = get_sbj_name(ssl_id, buf, sizeof(buf));
    if (0 < ercd) {
        buf[ercd] = '\0';
        shell_puts(ctrl, buf);
    } else {
        shell_printf(ctrl, SPL_LF" ... get_sbj_name() = %d" SPL_LF, ercd);
    }

    /* Issuer */
    shell_puts(ctrl, SPL_LF"  [Issuer ] ");
    ercd = get_issu_name(ssl_id, buf, sizeof(buf));
    if (0 < ercd) {
        buf[ercd] = '\0';
        shell_puts(ctrl, buf);
    } else {
        shell_printf(ctrl, SPL_LF" ... get_issu_name() = %d" SPL_LF, ercd);
    }

    /* DNS */
    shell_puts(ctrl, SPL_LF"  [DNS    ]");
    ercd = get_dns_cnt(ssl_id);
    if (0 <= ercd) {
        for (num = ercd; num; --num) {
            ercd = get_dns_name(ssl_id, buf, sizeof(buf), num - 1);
            shell_puts(ctrl, SPL_LF"    ");
            if (0 < ercd) {
                buf[ercd] = '\0';
                shell_puts(ctrl, buf);

                if (sizeof(buf) < ercd) {
                    shell_puts(ctrl, " (Buffer too short)");
                }
            } else {
                shell_printf(ctrl, SPL_LF" ... get_dns_name() = %d" SPL_LF, ercd);
            }
        }
    } else {
        shell_printf(ctrl, SPL_LF" ... get_dns_cnt() = %d" SPL_LF, ercd);
    }

    /* Verify the signature */
    shell_puts(ctrl,SPL_LF"  [Verify the signature]"SPL_LF "    ");
    ercd = get_sig_verify(ssl_id);
    shell_puts(ctrl, (ercd == E_OK) ? "OK" : "NG");

    /* Validity period */
    shell_puts(ctrl, SPL_LF"  [Validity period]"SPL_LF);
    shell_puts(ctrl, "    Start : ");
    ercd = get_valid_start(ssl_id, &tim);
    if (ercd == E_OK) {
        /* Print time */
        shell_printf(
            ctrl, 
            "%04d/%02d/%02d %02d:%02d:%02d"SPL_LF,  
            tim.year, 
            tim.mon, 
            tim.day, 
            tim.hour, 
            tim.min, 
            tim.sec
        );
    } else {
        shell_printf(ctrl, SPL_LF" ... get_valid_start() = %d" SPL_LF, ercd);
    }
    
    shell_puts(ctrl, "    End   : ");
    ercd = get_valid_end(ssl_id, &tim);
    if (ercd == E_OK) {
        /* Print time */
        shell_printf(
            ctrl, 
            "%04d/%02d/%02d %02d:%02d:%02d"SPL_LF,  
            tim.year, 
            tim.mon, 
            tim.day, 
            tim.hour, 
            tim.min, 
            tim.sec
        );
    } else {
        shell_printf(ctrl, SPL_LF" ... get_valid_end() = %d" SPL_LF, ercd);
    }
    
    shell_puts(ctrl, SPL_LF);
}
#endif


/* Command 'http_get' */
ER shell_usr_cmd_http_get(VP ctrl, INT argc, VB *argv[])
{
#if defined(ENA_AUTH_BASIC) || defined(ENA_AUTH_DIGEST)
    UB auth_flg = 0;
#endif
    ER ercd;
    T_HTTP_CLIENT hc;
    VB *url = argv[1];

    net_memset(&hc, 0, sizeof(hc));

    if (use_proxy) {
        hc.svr.ipa = prx_ipa;
        hc.svr.port = prx_port;
        hc.svr.ver = IP_VER4;
        hc.svr.num = 1;
        hc.prx = use_proxy;
    }
    else {
        ercd = http_get_ipaddr(url, &hc.svr, ID_SOC_DNSC, SPL_DNS_SERVER);
        if (ercd != E_OK) {
            return ercd;
        }
    }

    hc.http_sid = ID_SOC_HTTPC;
    hc.mpf_id = ID_HTTP_MSG_MPF;
    
#if defined(ENA_AUTH_BASIC) || defined(ENA_AUTH_DIGEST)
RETRY:
#endif
    ercd = http_cmd_get(&hc, url);
    if (ercd != E_OK) {
        http_cls(&hc);
        return ercd;
    }

    /* Display Response */
    shell_puts(ctrl, SPL_LF);
    ercd = rcv_contents_mng(ctrl, &hc);

    /* 401 */
    if (hc.response.code == 401) {
#if defined(ENA_AUTH_BASIC) || defined(ENA_AUTH_DIGEST)
        if (auth_flg) {
            /* Authentication failure */
            ercd = E_OBJ;
        }
#ifdef ENA_AUTH_BASIC
        else if (hc.authtype == WWW_AUTH_BASIC) {
            auth_flg = 1;
            hc.usr = basic_user;
            hc.pw = basic_pass;
            goto RETRY;
        }
#endif
#ifdef ENA_AUTH_DIGEST
        else if (hc.authtype == WWW_AUTH_DIGEST) {
            auth_flg = 1;
            hc.usr = digest_user;
            hc.pw = digest_pass;
            goto RETRY;
        }
#endif
#else
        /* Authentication failure */
        ercd = E_OBJ;
#endif
    }
#ifdef HTTPC_SSL_SUP
    else if (hc.response.code == 200) {
        if (hc.scm == HTTP_SCHEME_HTTPS) {
            shell_puts_SSL(ctrl, hc.ssl_id);
            
        }
    }
#endif

    http_cls(&hc);
    return ercd;
}

/* Command 'http_post' */
ER shell_usr_cmd_http_post(VP ctrl, INT argc, VB *argv[])
{
#if defined(ENA_AUTH_BASIC) || defined(ENA_AUTH_DIGEST)
    UB auth_flg = 0;
#endif
    ER ercd;
    T_HTTP_CLIENT hc;
    VB *url = argv[1];
    T_HTTP_POSTDATA data;

    net_memset(&hc, 0, sizeof(hc));
    ercd = http_get_ipaddr(url, &hc.svr, ID_SOC_DNSC, SPL_DNS_SERVER);
    if (ercd != E_OK) {
        return ercd;
    }
    hc.http_sid = ID_SOC_HTTPC;
    hc.mpf_id = ID_HTTP_MSG_MPF;

    net_memset(&data, 0, sizeof(data));
    data.buf = argv[2];
    data.buflen = net_strlen(data.buf);
    data.contenttype = argv[3];

#if defined(ENA_AUTH_BASIC) || defined(ENA_AUTH_DIGEST)
RETRY:
#endif
    ercd = http_cmd_post(&hc, argv[1], &data);
    if (ercd != E_OK) {
        http_cls(&hc);
        return ercd;
    }

    /* Display Response */
    shell_puts(ctrl, SPL_LF);
    ercd = rcv_contents_mng(ctrl, &hc);

    /* 401 */
    if (hc.response.code == 401) {
#if defined(ENA_AUTH_BASIC) || defined(ENA_AUTH_DIGEST)
        if (auth_flg) {
            /* Authentication failure */
            ercd = E_OBJ;
        }
#ifdef ENA_AUTH_BASIC
        else if (hc.authtype == WWW_AUTH_BASIC) {
            auth_flg = 1;
            hc.usr = basic_user;
            hc.pw = basic_pass;
            goto RETRY;
        }
#endif
#ifdef ENA_AUTH_DIGEST
        else if (hc.authtype == WWW_AUTH_DIGEST) {
            auth_flg = 1;
            hc.usr = digest_user;
            hc.pw = digest_pass;
            goto RETRY;
        }
#endif
#else
        /* Authentication failure */
        ercd = E_OBJ;
#endif
    }    
#ifdef HTTPC_SSL_SUP
    else if (hc.response.code == 200) {
        if (hc.scm == HTTP_SCHEME_HTTPS) {
            shell_puts_SSL(ctrl, hc.ssl_id);
            
        }
    }
#endif
    
    http_cls(&hc);
    return ercd;
}

/* 1MoCg̃t@CPOSTTv */
/* t@C1M_bytes.bin */
/*   ύXꍇFILE_SIZE̒`ύXĂB*/
/* t@C̓e͑S0x55 */
/* POSThttp://192.168.0.2/test/file_upload.php */
/*   ύXꍇURL_POST_FILE̕ύXĂB*/

#define FILE_SIZE (1024*1024)
#define URL_POST_FILE "http://192.168.0.2/test/file_upload.php"

ER shell_usr_cmd_http_post_file(VP ctrl, INT argc, VB *argv[])
{
    T_HTTP_CLIENT hc;
    T_HTTP_POSTDATA data;
    VB *url = URL_POST_FILE;
    ER ercd;
    UINT i;
    UINT len;
    
    net_memset(&hc, 0, sizeof(hc));
    
    /* URLT[o[IPAhX擾 */
    /* HTTPNCAgŎgpf[^(T_HTTP_CLIENT^)̃oϐsvr֊i[ */
    ercd = http_get_ipaddr(
               url,            /* URL */
               &hc.svr,        /* IPAhXi[ꏊ */
               ID_SOC_DNSC,    /* DNSNCAgŎgp\Pbg̃\PbgID */
               SPL_DNS_SERVER  /* DNST[oIPAhX */
           );
    if (ercd != E_OK) {
        return ercd;
    }
    
    hc.http_sid = ID_SOC_HTTPC;     /* \PbgID */
    hc.mpf_id   = ID_HTTP_MSG_MPF;  /* Œ蒷v[ID */
    
    /* 
     * http_cmd_post()̒POSTf[^𑗐MȂ̂ŁA
     * flagHTTPC_FLG_SND_BODY_LATERrbgZbgĂ
     */
    hc.flag = HTTPC_FLG_SND_BODY_LATER;
    
    /*
     * POST̃f[^̓t@C̓eȊÕf[^(t@CAwb_)
     * 
     * ȉMf[^
     *
     *    --BOUNDARY\r\n
     *    Content-Dispsition: form-data; name="userfile"; filename="1M_bytes.bin"\r\n
     *    Content-Type: application/octet-stream\r\n
     *    \r\n
     *    t@C̓e
     *    \r\n
     *    --BOUNDARY--\r\n
     * ܂őMf[^
     * L̃oCgContent-Lengthwb_փZbgđM邽߂ɁA
     * data.buflen֏L̃f[^̃oCgZbgB
     */
    
    net_memset(&data, 0, sizeof(data));
    data.buf = NULL;
    
    net_strcpy(http_buf, 
               "--BOUNDARY\r\n"
               "Content-Disposition: form-data; name=\"userfile\"; filename=\"1M_bytes.bin\"\r\n"
               "Content-Type: application/octet-stream\r\n"
               "\r\n"
    );
    
    data.buflen = net_strlen(http_buf);
    data.buflen += FILE_SIZE;
    data.buflen += sizeof("\r\n"
                          "--BOUNDARY--\r\n") - 1;
                         
    data.contenttype = "multipart/form-data; boundary=BOUNDARY";
    
    /* PostNGXg𑗐M (http_cmd_post̏ŃT[oƐڑ)
     * flagHTTPC_FLG_SND_BODY_LATERrbgZbgĂ̂ŁA
     * Rec̑MƃX|X̎M͍sȂ
     */
    ercd = http_cmd_post(&hc, url, &data);
    
    /* RecM (t@C̓e̒O܂) */
    ercd = http_snd(&hc, (UB*)http_buf, net_strlen(http_buf));
    
    /* t@C̓e𑗐M */
    /* f[^TCY傫̂ŁAhttp_snd()s */
    net_memset(http_buf, 0x55, sizeof(http_buf));
    
    for (i = 0; i < FILE_SIZE; i += len) {
        len = FILE_SIZE - i;
        if (len > sizeof(http_buf)) {
            len = sizeof(http_buf);
        }
        
        ercd = http_snd(&hc, (UB *)http_buf, len);
    }
    
    /* t@C̓ě̃f[^𑗐M
     *    \r\n
     *    --BOUNDARY--\r\n
     */
    net_strcpy(http_buf, 
               "\r\n"
               "--BOUNDARY--\r\n");
    ercd = http_snd(&hc, (UB *)http_buf, net_strlen(http_buf));
    
    /* X|X(Xe[^XCƃwb_)M */
    ercd = http_rcv_status(&hc);
    
    /* T[o瑗R[h\ */
    shell_printf(ctrl, "\r\n");
    shell_printf(ctrl, "code = %d\r\n", hc.response.code);
    
    /* T[o瑗MRecM */
    /* M֐̒g̓TvR[h̃t@CQƁB
     */
    if (hc.response.chunked) {
        ercd = rcv_chunked_contents(ctrl, &hc);
    } else if (hc.response.contentlen) {
        ercd = rcv_contents(ctrl, &hc);
    } else {
        ercd = rcv_continuous_contents(ctrl, &hc);
    }
    
    /* ؒf */
    http_cls(&hc);
    
    return ercd;
}

/* Command 'http_cfg' */
ER shell_usr_cmd_http_cfg(VP ctrl, INT argc, VB *argv[])
{
#define SPL_HTTP_BUF    ((SPL_HTTP_USR_BUF > SPL_HTTP_PW_BUF) ? SPL_HTTP_USR_BUF : SPL_HTTP_PW_BUF)
    VB tmp[SPL_HTTP_BUF];

#ifdef ENA_AUTH_BASIC
    shell_puts(ctrl, SPL_LF" Basic UserName: ");
    shell_gets(ctrl, tmp, sizeof(tmp));
    if (0 < net_strlen(tmp))    net_strcpy(basic_user, tmp);

    shell_puts(ctrl, SPL_LF" Basic Password: ");
    shell_gets(ctrl, tmp, sizeof(tmp));
    if (0 < net_strlen(tmp))    net_strcpy(basic_pass, tmp);
#endif

#ifdef ENA_AUTH_DIGEST
    shell_puts(ctrl, SPL_LF" Digest UserName: ");
    shell_gets(ctrl, tmp, sizeof(tmp));
    if (0 < net_strlen(tmp))    net_strcpy(digest_user, tmp);

    shell_puts(ctrl, SPL_LF" Digest Password: ");
    shell_gets(ctrl, tmp, sizeof(tmp));
    if (0 < net_strlen(tmp))    net_strcpy(digest_pass, tmp);
#endif

    /* Proxy Settings (no error check) */
    shell_puts(ctrl, SPL_LF" Use Proxy (0:No, 0<>:Yes): ");
    shell_gets(ctrl, tmp, sizeof(tmp));
    if (0 < net_strlen(tmp))    use_proxy = net_atoi(tmp);
    if (use_proxy) {
        shell_puts(ctrl, SPL_LF"  Proxy Server IP  : ");
        shell_gets(ctrl, tmp, sizeof(tmp));
        if (0 < net_strlen(tmp))    prx_ipa = ip_aton(tmp);

        shell_puts(ctrl, SPL_LF"  Proxy Server Port: ");
        shell_gets(ctrl, tmp, sizeof(tmp));
        if (0 < net_strlen(tmp))    prx_port = net_atoi(tmp);
    }

    return E_OK;
}
